#include "TimerTaskScheduler.h"
#include "TimerTask.h"
#include "List.h"

class TimerTaskEntry {
    friend class TimerTaskScheduler;
public:    
    TimerTaskEntry(TimerTask * _task, long _nextTime, long _interval) {
        task = _task;
        nextTimeMS = _nextTime;
        intervalMS = _interval;
    }

    long getIntervalMS() const {
        return intervalMS;
    }

    long getNextTimeMS() const {
        return nextTimeMS;
    }

    TimerTask* getTask() const {
        return task;
    }
    
    long getTaskId() const {
        return task->getTaskId();
    }
    
private:
    void setNextTime() {
        if (intervalMS != 0) {
            nextTimeMS = nextTimeMS + intervalMS;
        } else {
            nextTimeMS = 0;
        }
    }
    
    TimerTask * task;
    
    long nextTimeMS;
    long intervalMS;
    
};

long TimerTaskScheduler::currentTaskId = 0;

TimerTaskScheduler::TimerTaskScheduler(byte _pin, long _resolution) 
        : Timer(_pin) {
    resolution = _resolution;
    taskMap = new Map<long, TimerTaskEntry *>();
    processTime = 0;
}

void TimerTaskScheduler::scheduleTask(TimerTask * task, long firstInterval, long interval) {
    removeTask(task);
    currentTaskId++;
    task->setTaskId(currentTaskId);
    TimerTaskEntry * entry = new TimerTaskEntry(task, millis() + firstInterval, interval);
    taskMap->add(currentTaskId, entry);
}

void TimerTaskScheduler::scheduleTask(TimerTask * task, long interval) {
    scheduleTask(task, interval, interval);
}

void TimerTaskScheduler::scheduleTaskSingle(TimerTask * task, long interval){
    scheduleTask(task, interval, 0);
}

void TimerTaskScheduler::removeTask(TimerTask * task) {
    if (taskMap->containsKey(task->getTaskId())) {
        taskMap->remove(task->getTaskId());
    }
}

long TimerTaskScheduler::getProcessTime() {
    return processTime;
}

void TimerTaskScheduler::process(long k, TimerTaskEntry * entry, void * data) {
    TimerTaskScheduler * scheduler = (TimerTaskScheduler *)data;
    if (scheduler->getProcessTime() >= entry->getNextTimeMS()) {
        entry->getTask()->timerExecute();
        entry->setNextTime();
    }
}

void TimerTaskScheduler::addToDelete(long k, TimerTaskEntry * entry, void * data) {
    List<long> * deleteKeys = (List<long> *)data;
    if (entry->getNextTimeMS() == 0) {
        deleteKeys->add(k);
    } else if (entry->getTask()->isTerminated()) {
        deleteKeys->add(k);
    }
}

void TimerTaskScheduler::deleteEntry(long k, void * data) {
    TimerTaskScheduler * scheduler = (TimerTaskScheduler *)data;
    scheduler->taskMap->remove(k);
}

void TimerTaskScheduler::handleInterrupt(byte intPin) {
    if (isOk()) {
        Timer::handleInterrupt(intPin);
        // See if we have reached resolution change in time
        if ((getLastTickTimeMS() - processTime) > resolution) {
            processTime = getLastTickTimeMS();
            // Process all tasks that have reached their time
            taskMap->forAll(process, this);
            // Remove any tasks that have a zero next time - i.e. one time only tasks
            List<long> * deleteKeys = new List<long>();
            taskMap->forAll(addToDelete, deleteKeys);
            deleteKeys->forAll(deleteEntry, this);
            delete deleteKeys;
        }
    }
}

void TimerTaskScheduler::onAttach(byte intPin) {
    Timer::onAttach(intPin);
    if (Timer::isOk()) {
        processTime = millis();
    }
}

void TimerTaskScheduler::onDetach(byte intPin) {
    Timer::onDetach(intPin);
    processTime = 0;
}

